/*
    sptrace.c - Secure ptrace() module

    Copyright (C) 2003, 2004, 2007, 2009  Krzysztof Burghardt <krzysztof@burghardt.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
   Version Information
*/
#define VERSION "1.4.2"
#define AUTHOR  "Krzysztof Burghardt <krzysztof@burghardt.pl>"
#define DESC    "Secure ptrace() Linux Kernel Module (LKM)"
#define LICENSE "GPL"

#ifndef __KERNEL__
# define __KERNEL__
#endif

#ifndef MODULE
# define MODULE
#endif

/*
   Includes
*/
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
# include <linux/modversions.h>
# include <linux/config.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
# include <sys/syscall.h>
#else
# include <linux/unistd.h>
#endif 

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 extern void **sys_call_table;
#else
 /* Please remove or comment out next line. */
# error You did not RTFM carefully enough !!!
 /* Determine sys_call_table address,
  * e.g.: grep sys_call_table /boot/System.map
  * and fill in blank below */
# define SYS_CALL_TABLE_ADDR 0x????????
 void **sys_call_table;
#endif

/*
   Global Variables
*/

/* GID of group allowed to use ptrace() call */
static gid_t ptrace_group = 0;

/* Pointer to original ptrace call */
static asmlinkage long (*orig_ptrace) (long, long, long, long);

/*
   Helper Function
*/
static inline void
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
alert(const task_t * who, const char *message)
#else
alert(const struct task_struct *who, const char *message)
#endif
{
    static pid_t last_pid = 0;

    if (last_pid != who->pid) {
        last_pid = who->pid;

        printk("sptrace: %s (%s:%d) UID(%d) EUID(%d), "
             "parent (%s:%d) UID(%d) EUID(%d)\n", message,
#   if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
             who->comm, who->pid, who->uid, who->euid,
             who->p_pptr->comm, who->p_pptr->pid, who->p_pptr->uid,
             who->p_pptr->euid);
#   else
             who->comm, who->pid, who->uid, who->euid,
             who->parent->comm, who->parent->pid,
             who->parent->uid, who->parent->euid);
#   endif
    }

    return;

}

/*
   Security through obscurity
*/
static asmlinkage long
sptrace(long request, long pid, long addr, long data)
{
    if (in_egroup_p(ptrace_group) == 1) {
        alert(current, "ptrace() by");
        return (orig_ptrace) (request, pid, addr, data);
    }

    alert(current, "ptrace() DENIED for");

    return -EPERM;
}

/*
   init function
*/
static int __init
sptrace_init_module(void)
{
#   if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
    sys_call_table = (void **) SYS_CALL_TABLE_ADDR;
#   endif
    printk("sptrace: secure ptrace() module v%s (ptrace group id: %ld)\n",
        VERSION, (long)ptrace_group);
    orig_ptrace = sys_call_table[__NR_ptrace];
    sys_call_table[__NR_ptrace] = sptrace;
    return 0;
}

/*
   exit function
*/
static void __exit
sptrace_cleanup_module(void)
{
    printk("sptrace: unloading secure ptrace() module\n");
    sys_call_table[__NR_ptrace] = orig_ptrace;
    return;
}

/*
   module_init & module_exit
*/
module_init(sptrace_init_module);
module_exit(sptrace_cleanup_module);

/*
   Author and description
*/

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESC);
MODULE_LICENSE(LICENSE);

/*
   Parameters
*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(ptrace_group, "i");
#else
module_param(ptrace_group, int, 400);
#endif
MODULE_PARM_DESC(ptrace_group,
         "GID of group allowed to ptrace (default: 0, -1 to disable even for root)");
/*
   EOF
*/
